#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Acyclic CSDF Graph Model'

import sys
import math
import os.path

from ActorModel import ActorModel
from ChannelModel import ChannelModel
import Utilities


verbose = False


class ACSDFModel:
	"""ACSDFModel class 
	
	Represents an acyclic Cyclo-Static DataFlow (CSDF) graph.
	CSDF is originally defined in the following article:
	G. Bilsen, M. Engels, R. Lauwereins, and J. Peperstraete, "Cycle-static dataflow," IEEE Transactions on Signal Processing, vol. 44, no. 2, pp. 397-408, February 1996.
	"""
	
	
	# Attributes related to parsing the SPS file
	ACTOR_FIELD_STR = "A"
	ACTOR_FIELD_LENGTH = 5
	ACTOR_FIELD_ID_INDEX = 1
	ACTOR_FIELD_NAME_INDEX = 2
	ACTOR_FIELD_FUNCTION_INDEX = 3
	ACTOR_FIELD_EXECUTION_INDEX = 4
	ACTOR_FIELD_REPETITION_INDEX = 5
	CHANNEL_FIELD_STR = "C"
	CHANNEL_FIELD_LENGTH = 7
	CHANNEL_FIELD_ID_INDEX = 1
	CHANNEL_FIELD_NAME_INDEX = 2
	CHANNEL_FIELD_SOURCE_INDEX = 3
	CHANNEL_FIELD_DESTINATION_INDEX = 4
	CHANNEL_FIELD_PROD_SEQUENCE_INDEX = 5
	CHANNEL_FIELD_CONS_SEQUENCE_INDEX = 6
	
	# Deadline constraint type	
	D1, D2, D3, D4, D5 = range(1,6)
	
	# The supported start time and buffer computation methods.
	# IDP_SUF: Implicit-deadline with sufficient start time and buffer
	# IDP_MIN: Implicit-deadline with minimum start time and buffer
	# CDP_MIN: Constrained-deadline with minimum start time and buffer
	IDP_SUF, IDP_MIN, CDP_MIN = range(3)

	def __init__(self):
		"""The Constructor"""
		self.actors = []
		self.channels = []
		self.idp_latency = 0
		self.cdp_latency = 0
		self.alpha = 0
		self.num_of_levels = 0
		self.lcm = 0
		self.max_q_C = 0
		self.factor = 1
		self.method = None
		self.deadline_type = -1
		self.max_token_size = 0
	
		
	def get_max_token_size(self):
		return self.max_token_size
		
	
	def set_max_token_size(self, token_size):
		self.max_token_size = token_size
	

	def add_actor(self, actor_id, name, function, wcet):
		"""Adds an actor to the graph"""
		am = ActorModel(self, actor_id, name, function, wcet)
		self.actors.append(am)
		

	def add_channel(self, channel_id, name, source, destination, production_sequence, consumption_sequence):
		"""Adds a channel to the graph"""
		cm = ChannelModel(channel_id, name, source, destination, production_sequence, consumption_sequence)
		self.channels.append(cm)
		for i in self.actors:
			if i.get_actor_id() == source:
				i.add_outport(channel_id)
			if i.get_actor_id() == destination:
				i.add_inport(channel_id)
				

	def get_actor(self, actor_id):
		"""Returns the actor object with the given actor_id. Returns None if no corresponding actor is found"""
		for i in self.actors:
			if (i.get_actor_id() == actor_id):
				return i
		return None
		

	def get_channel(self, channel_id):
		"""Returns the channel object with given channel_id. Returns None if no corresponding channel is found"""
		for i in self.channels:
			if (i.get_channel_id() == channel_id):
				return i
		return None
		

	def get_actors(self):
		"""Returns the set of actor objects"""
		return self.actors
		
	
	def get_sink_actor(self):
		levels_actors = self.find_levels()
		return levels_actors[-1]
		
	
	def get_channels(self):
		"""Returns the set of channel objects"""
		return self.channels
		

	def get_num_of_actors(self):
		"""Returns the number of actors in the graph"""
		return len(self.actors)
		

	def get_num_of_channels(self):
		""""Returns the number of channels in the graph"""
		return len(self.channels)
		

	def get_graph_name(self):
		"""Returns the graph name"""
		return self.name
		

	def set_graph_name(self, name):
		"""Set the graph name"""
		self.name = name
		
	
	def get_factor(self):
		return self.factor
		
	
	def set_factor(self, factor):
		assert type(factor) == int
		self.factor = factor


	def get_idp_latency(self):
		"""Returns the IDP latnecy"""
		return self.idp_latency
		

	def get_cdp_latency(self):
		"""Returns the CDP latency"""
		return self.cdp_latency		


	def update_parameters(self, interconnect_type, word_read_time, word_write_time):
		for a in self.get_actors():
			a.update_execution_time(interconnect_type, word_read_time, word_write_time)
		self.find_minimum_period_vector(self.get_factor())
		

	def find_minor_cycle(self, lb, major_cycle):
		minor_cycle = lb
		while minor_cycle <= major_cycle:
			if major_cycle % minor_cycle == 0:
				found = True
				for a in self.get_actors():
					val = minor_cycle + (minor_cycle - Utilities.gcd(minor_cycle, a.get_period()))
					if val > a.get_deadline():
						found = False
						break
				if found:
					return minor_cycle
			minor_cycle += 1
		return 0
	
	
	def compute_cyclic_executive_parameters(self):
		major_cycle = Utilities.lcmv(self.get_period_vector())
		lb = max(self.get_execution_vector())
		minor_cycle = self.find_minor_cycle(lb, major_cycle)
		if minor_cycle != 0:
		
			assert major_cycle % minor_cycle == 0
			for a in self.get_actors():
				val = minor_cycle + (minor_cycle - Utilities.gcd(minor_cycle, a.get_period()))
				assert val <= a.get_deadline()
				
			self.major_cycle = major_cycle
			self.minor_cycle = minor_cycle
		else:
			# Unable to find a minor cycle with no preemption
			self.major_cycle = major_cycle
			self.minor_cycle = minor_cycle					


	def find_levels(self):
		"""Finds the levels of a graph"""
		levels = list()
		actors = self.copy_actors()
		
		# TODO: Find a better way to check emptiness
		while len(actors) > 0: 
			working_set = set()
			for i in actors:
				if len(i.get_inports()) == 0:
					working_set.add(i)
					actors.remove(i)
			levels.append(working_set)

			for j in working_set:
				for l in j.get_outports():
					for k in actors:
						# TODO: Move the exception handling to the remove methods
						try:
							k.get_inports().index(l)
							k.remove_inport(l)
						except:
							pass
		return levels

			
	def find_minimum_period_vector(self, factor):
		"""Finds the minimum period vector and then multiplies it with factor"""
		max_q_C = 0
		for i in self.actors:
			val = i.get_wcet()*i.get_repetition() 
			if val > max_q_C:
				max_q_C = val
		self.max_q_C = max_q_C
		self.lcm = Utilities.lcmv(self.get_repetition_vector())
		self.alpha = self.lcm * int(math.ceil(float(self.max_q_C)/float(self.lcm))) * factor
		for i in self.actors:
			period = self.alpha / i.get_repetition()
			i.set_period(period)
			i.set_deadline(period) # Initial value of deadline
			assert i.get_period() * i.get_repetition() == self.alpha		

		if self.max_q_C % self.lcm == 0:
			self.is_matched_io_rates = True
		else:
			self.is_matched_io_rates = False

	
	def get_alpha(self):
		"""Returns the actor iteration period"""
		return self.alpha

			
	def get_matched_io_type(self):
		return self.is_matched_io_rates	
	

	def generate_cumulative_rates_idp(self):
		"""Generates the cumulative rates under IDP for all channels.
		These rates are used for computing prd and cns functions
		"""

		alpha = self.get_alpha()
		assert alpha > 0

		for c in self.get_channels():
			c.clear_cumulative_rates()
			src = self.get_actor(c.get_source())
			dst = self.get_actor(c.get_destination())
			prd_period = src.get_period()
			cns_period = dst.get_period()
			cns_repetition = dst.get_repetition()
			c.set_cumulative_prd_rates_idp(self.num_of_levels*alpha, prd_period)
			c.set_cumulative_cns_rates_idp(self.num_of_levels*alpha, cns_period, cns_repetition)
	
	
	def generate_cumulative_rates_cdp(self):
		"""Generates the cumulative rates under CDP for all channels.
		These rates are used for computing prd and cns functions
		"""

		alpha = self.get_alpha()
		assert alpha > 0
		
		for c in self.get_channels():
			c.clear_cumulative_rates()
			src = self.get_actor(c.get_source())
			dst = self.get_actor(c.get_destination())
			prd_period = src.get_period()
			prd_deadline = src.get_deadline()
			cns_period = dst.get_period()
			cns_repetition = dst.get_repetition()
			c.set_cumulative_prd_rates_cdp(self.num_of_levels*alpha, prd_period, prd_deadline)
			c.set_cumulative_cns_rates_cdp(self.num_of_levels*alpha, cns_period, cns_repetition)
										
	

	def find_earliest_start_time_vector_idp(self):
		"""Finds the earliest start time vector under the IDP model. Invoked when self.method = IDP_MIN"""

		levels = self.find_levels()
		level = 1
		
		for s in levels:
			for a in s:
				actor = self.get_actor(a.get_actor_id())
				assert actor.get_period() * actor.get_repetition() == self.alpha
				alpha = self.alpha
				if level == 1:
					actor.set_start_time(0)
				else:
					start_time_candidates = []
					for inport in actor.get_inports():
						channel = self.get_channel(inport)
						assert channel != None
						predecessor = self.get_actor(channel.get_source())
						assert predecessor != None
						
						if (actor.get_start_time() < (predecessor.get_start_time() + alpha) or actor.get_start_time() == 0):
							actor.set_start_time(predecessor.get_start_time() + alpha)
						
						t_i = predecessor.get_start_time()
						t_j = 0
						
						if actor.cns_idp(channel, t_j, t_i) > 0:
							t_j = t_i 
						
						gcd_val = Utilities.gcd(predecessor.get_period(), actor.get_period())
						#gcd_val = 1

						while t_j <= (predecessor.get_start_time() + alpha):
							found = True
							t_prime = max(t_i, t_j)
							for t_hat in range(0, int(alpha) + 1, int(actor.get_period())):
								prd_i = predecessor.prd_idp(channel, t_i , t_prime + t_hat)
								cns_j = actor.cns_idp(channel, t_j, t_prime + t_hat ) 
								net_tokens = prd_i - cns_j 
								if net_tokens < 0:
									found = False
									break
							if not found:
								t_j = t_j + gcd_val
							else:
								start_time_candidates.append(t_j)
								break
					if start_time_candidates != []:
						actor.set_start_time(max(start_time_candidates))
			level = level + 1
			
		self.compute_idp_latency()


	def find_earliest_start_time_cdp(self, actor, level):
		"""Finds the earliest start time of an actor under the CDP model. Invoked when self.method = CDP_MIN"""

		assert actor.get_period() * actor.get_repetition() == self.alpha
		alpha = self.alpha
		v_m = None
		if level == 1:
			start_time = 0
			actor.set_start_time(start_time)
		else:
			start_time = 0
			for inport in actor.get_inports():
				channel = self.get_channel(inport)
				assert channel != None
				predecessor = self.get_actor(channel.get_source())
				assert predecessor != None
				ps = predecessor.get_start_time() + alpha
				es = ps
				
				t_i = predecessor.get_start_time()
				t_j = 0
				
				#gcd_val = Utilities.gcdv([predecessor.get_deadline(), predecessor.get_period(), actor.get_period()])
				gcd_val = 1
				
				if t_j < t_i and actor.cns_cdp(channel, t_j, t_i) > 0:
					t_j = t_i
					
				while t_j <= (predecessor.get_start_time() + alpha):
					found = True
					t_prime = max(t_i, t_j)
					for t_hat in range(0, alpha + 1, actor.get_period()): 
						prd_i = predecessor.prd_cdp(channel, t_i , t_prime + t_hat)
						cns_j = actor.cns_cdp(channel, t_j, t_prime + t_hat )  
						net_tokens = prd_i - cns_j 
						if net_tokens < 0:
							found = False
							break
					if not found:
						t_j += gcd_val
					else:
						es = t_j
						break
				st = min(es,ps)
				
				if st > start_time:
					start_time = st
					v_m = predecessor
			if v_m != None:
				actor.set_start_time(start_time)
		return v_m

						
	def find_earliest_start_time_vector_cdp(self, deadline_type):
		"""Finds the earliest start time vector under the CDP model. Invoked when self.method = CDP_MIN"""
		
		levels = self.find_levels()
		level = 1
		for s in levels:
			for a in s:
				actor = self.get_actor(a.get_actor_id())
				v_m = None
				while True:
					v_m = self.find_earliest_start_time_cdp(actor, level)
					if v_m != None:
	
						if deadline_type == ACSDFModel.D1:
							v_m.set_deadline(v_m.get_period())
						elif deadline_type == ACSDFModel.D2:
							v_m.set_deadline(v_m.get_wcet() + (3*(v_m.get_period() - v_m.get_wcet())/4))
						elif deadline_type == ACSDFModel.D3:
							v_m.set_deadline(v_m.get_wcet() + ((v_m.get_period() - v_m.get_wcet())/2))
						elif deadline_type == ACSDFModel.D4:
							v_m.set_deadline(v_m.get_wcet() + ((v_m.get_period() - v_m.get_wcet())/4))
						elif deadline_type == ACSDFModel.D5:
							v_m.set_deadline(v_m.get_wcet()) 
							
						v_k = self.find_earliest_start_time_cdp(actor, level)
						assert v_k != None
						if v_k.get_actor_id() != v_m.get_actor_id():
							pass
						else:
							break
					else:
						break
			level = level + 1
		for a in self.get_actors():
			if a.get_outports() == []:
				if deadline_type == ACSDFModel.D1:
					a.set_deadline(a.get_period())
				elif deadline_type == ACSDFModel.D2:
					a.set_deadline(a.get_wcet() + (3*(a.get_period() - a.get_wcet())/4))
				elif deadline_type == ACSDFModel.D3:
					a.set_deadline(a.get_wcet() + ((a.get_period() - a.get_wcet())/2))
				elif deadline_type == ACSDFModel.D4:
					a.set_deadline(a.get_wcet() + ((a.get_period() - a.get_wcet())/4))
				elif deadline_type == ACSDFModel.D5:
					a.set_deadline(a.get_wcet()) 
		self.compute_cdp_latency()


	def get_paths(self,input_actor):
	
		assert input_actor.get_inports() == []
		done = False
		paths = [[] for i in range(len(input_actor.get_outports()))]
		actors = []   
		channels = []
		actors.append(input_actor)
		while not done:
			cactors = list(actors)
			for a in cactors:
				for c in a.get_outports():
					ch = self.get_channel(c)
					channels.append(ch) 
					actors.append(self.get_actor(ch.get_destination()))
					for p in paths:
						if p == [] or p[-1][2] == ch.get_source():
							p.append((ch.get_channel_id(), ch.get_source(), ch.get_destination()))
				if a.get_outports() != []:
					actors.remove(a)
			done = True
			for ac in actors:
				if ac.get_outports() != []:
					done = False
					break 
		return paths	


	def compute_idp_latency(self):
		"""Computes the IDP latency"""
		self.idp_latency = 0
		for a in self.get_actors():
			if a.get_outports() == []:
				for b in self.get_actors():
					if b.get_inports() == []:
						# Check if there is a path w_{a->b}
						paths = self.get_paths(b)
						for path in paths:
							if path[-1][2] == a.get_actor_id():
								# Bingo! Output path
								first_ch = self.get_channel(path[0][0])
								last_ch = self.get_channel(path[-1][0])
								g_i = ChannelModel.empty_firings(first_ch.get_production_sequence())
								f_j = ChannelModel.empty_firings(last_ch.get_consumption_sequence())
								input_actor = self.get_actor(path[0][1])
								output_actor = self.get_actor(path[-1][2])
								latency = output_actor.get_start_time() + output_actor.get_period()*(f_j + 1) - (input_actor.get_start_time() + g_i*input_actor.get_period())
								if latency > self.idp_latency:
									self.idp_latency = latency
	

	def compute_cdp_latency(self):
		"""Computes the CDP latency"""
		self.cdp_latency = 0
		for a in self.get_actors():
			if a.get_outports() == []:
				for b in self.get_actors():
					if b.get_inports() == []:
						# Check if there is a path w_{a->b}
						paths = self.get_paths(b)
						for path in paths:
							if path[-1][2] == a.get_actor_id():
								# Bingo! Output path
								first_ch = self.get_channel(path[0][0])
								last_ch = self.get_channel(path[-1][0])
								g_i = ChannelModel.empty_firings(first_ch.get_production_sequence())
								f_j = ChannelModel.empty_firings(last_ch.get_consumption_sequence())
								input_actor = self.get_actor(path[0][1])
								output_actor = self.get_actor(path[-1][2])
								latency = output_actor.get_start_time() + output_actor.get_period()*f_j + output_actor.get_deadline() - (input_actor.get_start_time() + g_i*input_actor.get_period())
								if latency > self.cdp_latency:
									self.cdp_latency = latency
				
						  
	def find_sufficient_start_time_vector(self):
		"""Finds the sufficient start time vector. Invoked when self.method = IDP_SUF"""
		levels = self.find_levels()
		level = 1
		for l in levels:
			for a in l:
				actor = self.get_actor(a.get_actor_id())
				actor.set_start_time((level - 1) * self.alpha)
			level = level + 1


	def find_minimum_buffer_sizes_idp(self):
		"""Finds the minimum buffer size vector under the IDP model. Invoked when self.method = IDP_MIN"""
		for ch in self.channels:
			source = self.get_actor(ch.get_source())
			destination = self.get_actor(ch.get_destination())
			buffer_size_candidates = list()
			buffer_size_candidates.append(1) # Minimum buffer size is 1
			t_max = max(source.get_start_time(), destination.get_start_time())
	 
			init_val = 0	   
			if source.get_start_time() <= destination.get_start_time():
				init_val = source.prd_cdp(ch, source.get_start_time(), destination.get_start_time())		
			if init_val > 0:
				buffer_size_candidates.append(init_val)
			gcd_val = Utilities.gcd(source.get_period(), destination.get_period())
			for k in range(0, int(self.alpha) + 1, int(gcd_val)):
				prd = source.prd_idp(ch, source.get_start_time(), t_max + k)
				cns = destination.cns_idp(ch, destination.get_start_time(), t_max + k)
				unconsumed = prd - cns
				if unconsumed > 0:
					buffer_size_candidates.append(unconsumed)
			min_buf_size = max(buffer_size_candidates)
			ch.set_buffer_size(min_buf_size )


	def find_minimum_buffer_sizes_cdp(self):
		"""Finds the minimum buffer size vector under the CDP model. Invoked when self.method = CDP_MIN"""
		for ch in self.channels:
			source = self.get_actor(ch.get_source())
			destination = self.get_actor(ch.get_destination())
			buffer_size_candidates = list()
			buffer_size_candidates.append(1)
			t_max = max(source.get_start_time(), destination.get_start_time())
	 
			init_val = 0	   
			if source.get_start_time() <= destination.get_start_time():
				init_val = source.prd_cdp(ch, source.get_start_time(), destination.get_start_time())		
			if init_val > 0:
				buffer_size_candidates.append(init_val)

			gcd_val = 1
			for k in range(0, self.alpha + 1, gcd_val):
				prd = source.prd_cdp(ch, source.get_start_time(), t_max + k)
				cns = destination.cns_cdp(ch, destination.get_start_time(), t_max + k)
				unconsumed = prd - cns
				if unconsumed > 0:
					buffer_size_candidates.append(unconsumed)
			min_buf_size = max(buffer_size_candidates)
			ch.set_buffer_size(min_buf_size )


	def find_sufficient_buffer_sizes(self):
		"""Finds the sufficient buffer size vector under the IDP model. Invoked when self.method = IDP_SUF"""
		for ch in self.channels:
			source = self.get_actor(ch.get_source())
			buffer_size = source.prd_idp(ch, source.get_start_time(), source.get_start_time() + self.get_alpha())
			ch.set_buffer_size(max(1, buffer_size))

	
	def clear(self):
		"""Clears the start time and buffer vectors of the graph"""
		for a in self.get_actors():
			a.set_start_time(0)
			a.set_deadline(a.get_period())
		for c in self.get_channels():
			c.set_buffer_size(sys.maxint)


	def get_execution_vector(self):
		"""Return the execution time vector"""
		execution_vector = list()
		for i in self.actors:
			execution_vector.append(i.get_wcet())
		return execution_vector
		
		
	def get_execution_vector_str(self):
		"""Return the execution time vector as a string"""
		execution_vector = "["
		for i in self.actors:
			execution_vector += i.get_name() + "_" + i.get_function() + " = " + str(i.get_wcet()) + ","
		execution_vector += "]"
		return execution_vector
		

	def get_period_vector(self):
		"""Return the period vector"""
		period_vector = list()
		for i in self.actors:
			period_vector.append(i.get_period())
		return period_vector


	def get_period_vector_str(self):
		"""Return the period vector as a string"""
		period_vector = "["
		for i in self.actors:
			period_vector += i.get_name() + "_" + i.get_function() + " = " + str(i.get_period()) + ","
		period_vector += "]"
		return period_vector		
	
	
	def get_buffer_size_vector(self):
		"""Return the buffer size vector"""
		buffer_vector = list()
		for c in self.channels:
			buffer_vector.append(c.get_buffer_size())
		return buffer_vector


	def get_buffer_size_vector_str(self):
		"""Return the buffer size vector as a string"""
		buffer_vector = "["
		for c in self.channels:
			buffer_vector += c.get_name() + " = " + str(c.get_buffer_size()) + ","
		buffer_vector += "]"
		return buffer_vector
	
	
	def get_start_time_vector(self):
		"""Return the start time vector"""
		start_time_vector = list()
		for i in self.actors:
			start_time_vector.append(i.get_start_time())
		return start_time_vector


	def get_start_time_vector_str(self):
		"""Return the start time vector as a string"""
		start_time_vector = "["
		for i in self.actors:
			start_time_vector += i.get_name() + "_" + i.get_function() + " = " + str(i.get_start_time()) + ","
		start_time_vector += "]"
		return start_time_vector


	def get_deadline_vector(self):
		"""Return the deadline vector"""
		deadline_vector = list()
		for i in self.actors:
			deadline_vector.append(i.get_deadline())
		return deadline_vector


	def get_deadline_vector_str(self):
		"""Return the deadline vector as a string"""
		deadline_vector = "["
		for i in self.actors:
			deadline_vector += i.get_name() + "_" + i.get_function() + " = " + str(i.get_deadline()) + ","
		deadline_vector += "]"
		return deadline_vector
	
	
	def get_repetition_vector(self):
		"""Return the repetition vecotr"""
		repetition_vector = list()
		for i in self.actors:
			repetition_vector.append(i.get_repetition())
		return repetition_vector
		

	def get_repetition_vector_str(self):
		"""Return the repetition vecotr as a string"""
		repetition_vector = "["
		for i in self.actors:
			repetition_vector += i.get_name() + "_" + i.get_function() + " = " + str(i.get_repetition()) + ","
		repetition_vector += "]"
		return repetition_vector


	def find_minimum_repetition_vector(self):
		"""Finds the basic repetition vector"""
		q_vec = []
		for i in self.actors:
			q_vec.append(i.get_repetition())
		gcd = Utilities.gcdv(q_vec)
		for i in self.actors:
			r = i.get_repetition()/gcd
			i.set_repetition(r)

			
	def get_major_cycle(self):
		return self.major_cycle
	
	
	def get_minor_cycle(self):
		return self.minor_cycle

			
	def get_utilization(self):
		"""Returns the total utilization"""
		utilization = 0.0
		for i in self.actors:
			utilization = utilization + float(i.get_wcet())/float(i.get_period())
		return utilization


	def get_density(self):
		"""Return the total density"""
		density = 0.0
		for i in self.actors:
			density = density + float(i.get_wcet())/min(float(i.get_period()), float(i.get_deadline()))
		return density
	
	
	def get_max_utilization(self):
		"""Returns the maximum utilization factor"""
		max_utilization = 0.0
		for i in self.actors:
			utilization = float(i.get_wcet())/float(i.get_period())
			if (utilization > max_utilization):
				max_utilization = utilization
		return max_utilization
	
	
	def copy_actors(self):
		"""Copies the actors list into a new one"""
		new_actors = list()
		for i in self.actors:
			a = ActorModel(self, i.actor_id, i.name, i.function, i.wcet)
			a.inports = list(i.inports)
			a.outports = list(i.outports)
			a.set_repetition(i.get_repetition())
			new_actors.append(a)
		return new_actors   


	@staticmethod
	def load_graph_from_file(filename, factor):
		"""Loads an ACSDF graph from an SPS file"""
		g = ACSDFModel()
		g.set_graph_name(os.path.basename(filename).split(".")[0])
		fp = open(filename, 'r')
		line = fp.readline()
		num_of_actors = int(line)
		assert num_of_actors > 0
		for i in range(0,num_of_actors):
			line = fp.readline()
			parts = line.split(" ")
			assert parts[0].upper() == ACSDFModel.ACTOR_FIELD_STR
			actor_id = int(parts[ACSDFModel.ACTOR_FIELD_ID_INDEX])
			name = parts[ACSDFModel.ACTOR_FIELD_NAME_INDEX]
			function = parts[ACSDFModel.ACTOR_FIELD_FUNCTION_INDEX]
			wcet = int(parts[ACSDFModel.ACTOR_FIELD_EXECUTION_INDEX])
			repetition = int(parts[ACSDFModel.ACTOR_FIELD_REPETITION_INDEX])
			g.add_actor(actor_id, name, function, wcet)
			g.get_actor(actor_id).set_repetition(repetition)
		line = fp.readline()
		num_of_channels = int(line)
		for i in range(0,num_of_channels):
			line = fp.readline()
			if line != "":
				prod_sequence = []
				cons_sequence = []
				assert line != None
				parts = line.split(" ")
				assert parts[0].upper() == ACSDFModel.CHANNEL_FIELD_STR
				channel_id = int(parts[ACSDFModel.CHANNEL_FIELD_ID_INDEX])
				name = parts[ACSDFModel.CHANNEL_FIELD_NAME_INDEX]
				source = int(parts[ACSDFModel.CHANNEL_FIELD_SOURCE_INDEX])
				destination = int(parts[ACSDFModel.CHANNEL_FIELD_DESTINATION_INDEX])
				prod_sequence_len = int(parts[ACSDFModel.CHANNEL_FIELD_PROD_SEQUENCE_INDEX])
				cons_sequence_len = int(parts[ACSDFModel.CHANNEL_FIELD_CONS_SEQUENCE_INDEX])
				for k in range(0,prod_sequence_len):
					prod_sequence.append(int(parts[ACSDFModel.CHANNEL_FIELD_LENGTH + k]))
				for k in range(0,cons_sequence_len):
					cons_sequence.append(int(parts[ACSDFModel.CHANNEL_FIELD_LENGTH + (prod_sequence_len) + k]))
				g.add_channel(channel_id, name, source, destination, prod_sequence, cons_sequence)

		g.num_of_levels = len(g.find_levels())
		if verbose:
			sys.stdout.write("Computing the minimum period vector...\n")
		g.find_minimum_repetition_vector()
		g.set_factor(factor)
		g.find_minimum_period_vector(g.get_factor())

		return g

	
	def calc_start_time_buffer_vector(self, method, deadline_type):
		"""Computes the start time and buffer size vector according to the given method"""
		assert method == self.IDP_SUF or method == self.IDP_MIN or method == self.CDP_MIN
		assert (deadline_type >= 0 and deadline_type <= ACSDFModel.D5) 
		
		# TODO: Fix this hack
		if self.method == None:
			self.method = method
		if self.deadline_type == -1:
			self.deadline_type = deadline_type
		
		if method == self.IDP_MIN:
			if verbose:
				sys.stdout.write("Computing the earliest start time vector [IDP - Minimal]...\n") 
			if ActorModel.RATES_IMPL_TYPE == ActorModel.CPU_MIN:
				self.generate_cumulative_rates_idp()
			self.find_earliest_start_time_vector_idp()	
			if verbose:
				sys.stdout.write("Computing the minimum buffer sizes [IDP - Minimal]...\n")		
			self.find_minimum_buffer_sizes_idp()
		if method == self.CDP_MIN:
			if verbose:
				sys.stdout.write("Computing the earliest start time vector [CDP (Type = " + str(deadline_type) + ") - Minimal]...\n")
			if ActorModel.RATES_IMPL_TYPE == ActorModel.CPU_MIN:			
				self.generate_cumulative_rates_cdp()
			self.find_earliest_start_time_vector_cdp(deadline_type)		
			if verbose:
				sys.stdout.write("Computing the minimum buffer sizes [CDP (Type = " + str(deadline_type) + ") - Minimal]...\n")
			self.find_minimum_buffer_sizes_cdp()
		if method == self.IDP_SUF:
			if verbose:
				sys.stdout.write("Computing the earliest start time vector [IDP - Sufficient]...\n")
			if ActorModel.RATES_IMPL_TYPE == ActorModel.CPU_MIN:			
				self.generate_cumulative_rates_idp()
			self.find_sufficient_start_time_vector()
			if verbose:
				sys.stdout.write("Computing the minimum buffer sizes [IDP - Sufficient]...\n")
			self.find_sufficient_buffer_sizes()

		self.compute_cyclic_executive_parameters()			
	
	
	@staticmethod
	def print_graph_dot(g, output_dir):
		graph = os.path.join(output_dir, g.get_graph_name() + ".dot")
		fp = open(graph, 'w')
		fp.write("digraph " + g.get_graph_name() + "{\n")
		fp.write("rankdir=TB;\n")	
		
		levels = g.find_levels()
		for s in levels:
			for a in s:
				actor = g.get_actor(a.get_actor_id())
				for outport in actor.get_outports():
					ch = g.get_channel(outport)
					successor = g.get_actor(ch.get_destination())
					fp.write(g.get_graph_name() + "_" + actor.get_name() + "_" + actor.get_function() + "[style=filled,color=\"lightblue\"];\n")
					fp.write(g.get_graph_name() + "_" + successor.get_name() + "_" + successor.get_function() + "[style=filled,color=\"lightblue\"];\n")
					fp.write(g.get_graph_name() + "_" + actor.get_name() + "_" + actor.get_function() + " -> " + g.get_graph_name() + "_" + successor.get_name() + "_" + successor.get_function() + ";\n")
		fp.write('label=\"' + g.get_graph_name() + '\";\n')
		fp.write("}\n")
		fp.flush()
		fp.close()

	
	def __repr__(self):
		s = ""
		for i in self.actors:
			s = s + repr(i) + "\n"
		for i in self.channels:
			s = s + repr(i) + "\n"
		return s
		

def print_usage():
	sys.stderr.write("""
Usage:
	> ACSDFModel.py method [deadline] filename
where method is:
	--idp-sufficient
	--idp-minimal
	--cdp-minimal\n
	
and deadline is the deadline assignment type (valid only with --cdp-minimal)
	1 : D_i = T_i
	2 : D_i = C_i + (T_i - C_i)*3/4
	3 : D_i = C_i + (T_i - C_i)/2
	4 : D_i = C_i + (T_i - C_i)/4
	5 : D_i = C_i	
""")	


def main():
	
	MAX_NUM_OF_PARAMETERS = 4
	MIN_NUM_OF_PARAMETERS = 3
	
	num_of_parameters = len(sys.argv)
	
	# Check the number of parameters
	if num_of_parameters < MIN_NUM_OF_PARAMETERS or num_of_parameters > MAX_NUM_OF_PARAMETERS:
		sys.stderr.write("ERROR: Incorrect number of parameters!\n")
		print_usage()
		sys.exit(-1)
	
	# Parse parameters
	given_method = sys.argv[1]
	if num_of_parameters == MAX_NUM_OF_PARAMETERS and given_method != "--cdp-minimal":
		sys.stderr.write("Invalid number of parameters!\n")
		print_usage()
		sys.exit(-1)
	elif num_of_parameters == MAX_NUM_OF_PARAMETERS and given_method == "--cdp-minimal":
		deadline_type = int(sys.argv[2])
		if deadline_type < ACSDFModel.D1 or deadline_type > ACSDFModel.D5:
			sys.stderr.write("Invalid deadline type!\n")
			print_usage()
			sys.exit(-1)
		filename = sys.argv[3]
		method = ACSDFModel.CDP_MIN
	elif given_method == "--idp-sufficient" or given_method == "--idp-minimal":
		filename = sys.argv[2]
		if given_method == "--idp-sufficient":
			method = ACSDFModel.IDP_SUF
		elif given_method == "--idp-minimal":
			method = ACSDFModel.IDP_MIN
		else:
			sys.stderr.write("ERROR: Invalid method!")
			print_usage()
			sys.exit(-1)
	else:
		sys.stderr.write("ERROR: Invalid method!")
		print_usage()
		sys.exit(-1)
	
	sys.stdout.write(" - Processing application " + filename + "...\n")
	g = ACSDFModel.load_graph_from_file(filename,1)

	# For benchmarking, print using minimal IDP first		
	g.calc_start_time_buffer_vector(ACSDFModel.IDP_MIN, 0)
	
	if verbose:
		sys.stdout.write(" -- Least common multiple of the repetitions (Q) = " + str(g.lcm) + "\n") 
		sys.stdout.write(" -- max(WCET_i * q_i) (eta) = " + str(g.max_q_C) + "\n") 
		sys.stdout.write(" -- Iteration period = " + str(g.alpha) + "\n")

		if g.get_matched_io_type():
			sys.stdout.write(" -- " + g.get_graph_name() + " is a matched I/O rates graph\n")
		else:
			sys.stdout.write(" -- " + g.get_graph_name() + " is a mis-matched I/O rates graph\n")

		if g.get_minor_cycle() != 0:
			sys.stdout.write(" -- Cyclic executive major cycle = " + str(g.get_major_cycle()) + ", Cyclic Executive Minor Cycle = " + str(g.get_minor_cycle()) + "\n")
		else:
			sys.stdout.write(" -- Unable to find a minor cycle with no preemptions!\n")

		sys.stdout.write(" -- Hyperperiod = " + str(Utilities.lcmv(g.get_period_vector())) + "\n")
		sys.stdout.write("\n")
		
	sys.stdout.write(" -- Number of actors = " + str(g.get_num_of_actors()) + "\n")
	sys.stdout.write(" -- Repetition vector = " + g.get_repetition_vector_str() + "\n")
	sys.stdout.write(" -- Worst-case execution time vector = " + g.get_execution_vector_str() + "\n")
	sys.stdout.write(" -- Minimum period vector = " + g.get_period_vector_str() + "\n")	
	sys.stdout.write(" -- Deadline vector = " + g.get_deadline_vector_str() + "\n")
	sys.stdout.write(" -- Start time vector = " + g.get_start_time_vector_str() + "\n")
	sys.stdout.write(" -- Buffer size vector = " + g.get_buffer_size_vector_str() + "\n")
	sys.stdout.write(" -- Total Utilization  = " + str(g.get_utilization()) + "\n")
	sys.stdout.write(" -- Graph maximum latency under the implicit-deadline model = " + str(g.get_idp_latency()) + "\n")
	
	ACSDFModel.print_graph_dot(g,os.getcwd())
	
	if given_method != "--idp-minimal":
		g.clear()

		if given_method == "--cdp-minimal":
			g.calc_start_time_buffer_vector(method, deadline_type)
		else:
			g.calc_start_time_buffer_vector(method, 0)
		sys.stdout.write("\n")

		sys.stdout.write(" - Updating the graph parameters assuming the constrained-deadline model...\n")		
		sys.stdout.write(" -- Deadline vector = " + g.get_deadline_vector_str() + "\n")
		sys.stdout.write(" -- Start time vector = " + g.get_start_time_vector_str() + "\n")
		sys.stdout.write(" -- Buffer size vector = " + g.get_buffer_size_vector_str() + "\n")
		sys.stdout.write(" -- Total Utilization  = " + str(g.get_utilization()) + "\n")
		sys.stdout.write(" -- Total Density = " + str(g.get_density()) + "\n")	
		sys.stdout.write(" -- Graph maximum latency under the implicit-deadline model = " + str(g.get_idp_latency()) + "\n")
		sys.stdout.write(" -- Graph maximum latency under the constrained-deadline model = " + str(g.get_cdp_latency()) + "\n")

if __name__ == "__main__":
	main()

