#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Actor Model for CSDF Graph'


import math

import PlatformParameters
import Utilities

class ActorModel:
	"""ActorModel class
	Represents an actor in a CSDF graph and the corresponding real-time task
	"""
	
	"""
	The supported implementations for computing prd and cns functions. Stored in self.RATES_IMPL_TYPE
	CPU_MIN: Pre-computes the cumulative prd. and cns. rates. Less CPU time but needs more memory
	MEM_MIN: Computes the the cumulative prd. and cns. rates on-the-fly. More CPU time but less memory
	Default is MEM_MIN
	"""
	CPU_MIN, MEM_MIN = range(2)
	
	RATES_IMPL_TYPE = MEM_MIN

	def __init__(self, graph, actor_id, name, function, wcet):
		self.graph = graph
		self.actor_id = actor_id
		self.task_id = actor_id
		self.name = name
		self.function = function
		self.wcet = wcet
		self.io_time = 0
		self.period = 0
		self.start_time = 0
		self.deadline = 0
		self.utilization = 0.0
		self.density = 0.0
		self.priority = 0
		self.P = 0
		self.contention = 0
		self.affinity = -1
		self.inports = []
		self.outports = []
		
		if self.RATES_IMPL_TYPE == self.CPU_MIN:
			self.prd_idp = self.prd_idp_cpu_min
			self.cns_idp = self.cns_idp_cpu_min
			self.prd_cdp = self.prd_cdp_cpu_min
			self.cns_cdp = self.cns_cdp_cpu_min			
		elif self.RATES_IMPL_TYPE == self.MEM_MIN:
			self.prd_idp = self.prd_idp_mem_min
			self.cns_idp = self.cns_idp_mem_min
			self.prd_cdp = self.prd_cdp_mem_min
			self.cns_cdp = self.cns_cdp_mem_min			
		
		
	def get_actor_id(self):
		return self.actor_id
		
		
	def get_task_id(self):
		return self.task_id
		
		
	def get_graph(self):
		return self.graph
		
				
	def get_name(self):
		return self.name
		
	
	def get_full_name(self):
		return self.graph.get_graph_name() + "_" + self.name


	def get_function(self):
		return self.function
		
		
	def get_start_time(self):
		return self.start_time
		
		
	def get_inports(self):
		return self.inports
		
		
	def get_outports(self):
		return self.outports
		
		
	def get_period(self):
		return self.period
		
		
	def get_repetition(self):
		return self.repetition
		
		
	def get_wcet(self):
		if PlatformParameters.TIME_IN_CYCLES:
			time_in_cycles = self.wcet + self.io_time
			return time_in_cycles
		else:
			time_in_ticks = int(math.ceil(float(self.wcet + self.io_time) / float(PlatformParameters.CYCLES_IN_TICK)))
			return time_in_ticks				


	def get_deadline(self):
		return self.deadline

	
	def get_priority(self):
		return self.priority

	
	def get_utilization(self):
		self.utilization = float(self.get_wcet())/float(self.get_period())
		return self.utilization

	
	def get_density(self):
		self.density = float(self.get_wcet())/min(float(self.get_period()),float(self.get_deadline()))
		return self.density

	
	def get_P(self):
		if self.P == 0:
			l = []
			for i in self.get_inports():
				c = self.graph.get_channel(i)
				l.append(len(c.get_consumption_sequence()))
			for i in self.get_outports():
				c = self.graph.get_channel(i)
				l.append(len(c.get_production_sequence()))
			self.P = Utilities.lcmv(l)
		return int(self.P)

	
	# TODO: See if we can compute the numbers of tokens once and store them
	def update_execution_time(self, interconnect_type, word_read_time, word_write_time):
		P = self.get_P()
		total_time = 0
		for k in range(P):
			read_tokens = 0
			written_tokens = 0		
			for i in self.get_inports():
				c = self.graph.get_channel(i)
				cs = c.get_consumption_sequence()
				read_tokens += cs[k % len(cs)]
			
			for i in self.get_outports():
				c = self.graph.get_channel(i)
				ps = c.get_production_sequence()
				written_tokens += ps[k % len(ps)]
			
			if interconnect_type == PlatformParameters.P2P:
				read_time = read_tokens * self.graph.get_max_token_size() * word_read_time
				write_time = written_tokens * self.graph.get_max_token_size() * word_write_time
				total = read_time + write_time
				if total > total_time:
					total_time = total
			elif interconnect_type == PlatformParameters.AXI_CB:
				read_time = read_tokens * self.graph.get_max_token_size() * (self.get_contention() + 1) * word_read_time
				# No contention in writes on AXI crossbar
				write_time = written_tokens * self.graph.get_max_token_size() * word_write_time
				total = read_time + write_time
				if total > total_time:
					total_time = total
			else:
				assert False
		if total_time > 0:
			self.io_time = total_time
		else:
			self.io_time = 0

	
	def get_contention(self):
		return self.contention
	

	def get_affinity(self):
		return self.affinity
		
	
	def set_affinity(self, affinity):
		self.affinity = affinity
	
	
	def set_contention(self, contention):
		self.contention = contention
		

	def set_task_id(self, task_id):
		self.task_id = task_id
		
	
	def set_priority(self, priority):
		self.priority = priority
		
		
	def set_start_time(self, start_time):
		self.start_time = start_time
		
		
	def set_period(self, period):
		self.period = period
		
		
	def set_repetition(self, repetition):
		self.repetition = repetition
		
		
	def set_wcet(self, wcet):
		self.wcet = wcet


	def set_deadline(self, deadline):
		self.deadline = deadline
		

	def add_inport(self, channel_id):
		self.inports.append(channel_id)


	def add_outport(self, channel_id):
		self.outports.append(channel_id)


	def remove_inport(self, channel_id):
		self.inports.remove(channel_id)


	def remove_outport(self, channel_id):
		self.outports.remove(channel_id)


	def prd_cdp_mem_min(self, channel, start_time, end_time): 
		"""Returns the cumulative number of produced tokens under the CDP model in the interval [start_time, end_time) using the MEM_MIN approach"""
		
		if end_time <= start_time:
			return 0
		firings = int(math.floor(float(end_time - start_time)/float(self.get_period())))			
		if ( float(((end_time - start_time) % self.get_period()))/float(self.get_deadline())) >= 1.0 :
			firings = firings + 1
			
		produced_tokens = 0
		prd_seq = channel.get_production_sequence()
		prd_seq_len = len(prd_seq)
		for k in range(0,firings,1):
			produced_tokens = produced_tokens + prd_seq[k % prd_seq_len]
		return produced_tokens	
	

	def cns_cdp_mem_min(self, channel,  start_time, end_time): 
		"""Returns the cumulative number of consumed tokens under the CDP model in the interval [start_time, end_time] using the MEM_MIN approach"""
		if end_time < start_time:
			return 0
		consumed_tokens = self.cns_idp_mem_min(channel, start_time, end_time)
		return consumed_tokens


	def prd_idp_mem_min(self, channel, start_time, end_time):
		"""Returns the cumulative number of produced tokens under the IDP model in the interval [start_time, end_time) using the MEM_MIN approach"""
		if end_time <= start_time:
			return 0
		firings = int(math.floor(float(end_time - start_time)/float(self.get_period())))
		produced_tokens = 0
		prd_seq = channel.get_production_sequence()
		prd_seq_len = len(prd_seq)
		for k in range(0,firings,1):
			produced_tokens = produced_tokens + prd_seq[k % prd_seq_len]
		return produced_tokens	
	

	def cns_idp_mem_min(self, channel,  start_time, end_time):
		"""Returns the cumulative number of consumed tokens under the IDP model in the interval [start_time, end_time] using the MEM_MIN approach"""
		if end_time < start_time:
			return 0
		firings = int(math.ceil(float((end_time - start_time))/float(self.get_period())))
		if ((end_time - start_time) % self.get_period() == 0 and firings < self.get_repetition()):
			firings = firings + 1

		consumed_tokens = 0
		cns_seq = channel.get_consumption_sequence()
		cns_seq_len = len(cns_seq)
		for k in range(0,firings,1):
			consumed_tokens = consumed_tokens + cns_seq[k % cns_seq_len]
		return consumed_tokens
	

	def prd_idp_cpu_min(self, channel, start_time, end_time):
		"""Returns the cumulative number of produced tokens under the IDP model in the interval [start_time, end_time) using the CPU_MIN approach"""
		if end_time <= start_time:
			return 0
		firings = int(math.floor(float(end_time - start_time)/float(self.get_period())))
		return channel.get_cumulative_prd_rates()[firings]
	

	def cns_idp_cpu_min(self, channel,  start_time, end_time):
		"""Returns the cumulative number of consumed tokens under the IDP model in the interval [start_time, end_time] using the CPU_MIN approach"""
		if end_time < start_time:
			return 0
		firings = int(math.ceil(float((end_time - start_time))/float(self.get_period())))
		if ((end_time - start_time) % self.get_period() == 0 and firings < self.get_repetition()):
			firings = firings + 1
		return channel.get_cumulative_cns_rates()[firings-1]
	

	def prd_cdp_cpu_min(self, channel, start_time, end_time):
		"""Returns the cumulative number of produced tokens under the CDP model in the interval [start_time, end_time) using the CPU_MIN approach"""
		if end_time <= start_time:
			return 0
		firings = int(math.floor(float(end_time - start_time)/float(self.get_period())))			
		if ( float(((end_time - start_time) % self.get_period()))/float(self.get_deadline())) >= 1.0 :
			firings = firings + 1
		return channel.get_cumulative_prd_rates()[firings]
		

	def cns_cdp_cpu_min(self, channel,  start_time, end_time):
		"""Returns the cumulative number of consumed tokens under the CDP model in the interval [start_time, end_time] using the CPU_MIN approach"""
		if end_time < start_time:
			return 0
		firings = int(math.ceil(float((end_time - start_time))/float(self.get_period())))
		if ((end_time - start_time) % self.get_period() == 0 and firings < self.get_repetition()):
			firings = firings + 1
		return channel.get_cumulative_cns_rates()[firings-1]
	

	def __repr__(self):
		s = self.graph.get_graph_name() + "_" + self.name + "_" + self.function
		#s += "Actor Id = " + str(self.actor_id) + ", name = " + self.name + ", C = " + str(self.wcet) + ", T = " + str(self.period) + ", D = " + str(self.deadline) + ", S = " + str(self.start_time) + ", q = " + str(self.repetition)
		#s += ", Inports = "
		#for i in self.inports:
		#	s += str(i) + ", "
		#s += ", Outports = "
		#for i in self.outports:
		#	s += str(i) + ", "
		return s

