#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Converts a Cyclo-Static Dataflow graph in the .gph format to SDF^3 XML and SPS formats'

import os
import sys
import tempfile
import subprocess

ENCODING = 'UTF-8'

class Port:
	"""Represents a Port in the input graph"""

	def __init__(self, actor, name, p_type, p_id, rate):
		self.actor = actor
		self.name = name
		self.rate = rate
		self.p_type = p_type
		self.p_id = p_id


	def print_port(self):
		print("Port " + self.name + " of Actor " + self.actor + " with rate " + self.rate)



class Node:
	"""Represents a Node in the input graph"""

	def find_wcet(self):
		m = 0
		for w in self.wcet:
			x = int(w)
			if x > m:
				m = x
		return m
	

	def __init__(self, name, function, node_id, seq_length, wcet, in_ports, out_ports):
		self.name = name
		self.function = function
		self.node_id = node_id
		self.wcet = wcet
		self.wcet_val = self.find_wcet()
		self.seq_length = seq_length
		self.inports = in_ports
		self.outports = out_ports
	def print_node(self):
		print("Node " + self.name + "_" + self.function + ", ID = " + self.node_id + ", WCET = " + self.wcet_val)

	
		
class Edge:
	"""Represents an Edge in the input graph"""

	def __init__(self, name, edge_id, src_actor, dst_actor, src_port, dst_port):
		self.name = name
		self.edge_id = edge_id
		self.src_actor = src_actor
		self.dst_actor = dst_actor
		self.src_port = src_port
		self.dst_port = dst_port


	def print_edge(self):
		print("Edge " + self.edge_id + " connecting src_actor = " + self.src_actor + " and dst_actor = " + self.dst_actor)


		
class CSDFConverter:
	"""Converts a CSDF .gph graph to either SDF^3 XML format or SPS format"""

	def __init__(self, input_file, output_file, sdf3analysis_path):
		"""Constructor"""
		
		self.input_file = input_file
		self.sdf3analysis_path = sdf3analysis_path
		self.output_file = output_file


	def sdf3_init_file(self, name):
		"""Open and Initializes the SDF3 XML file"""
		
		self.output_file.write("""<?xml version="1.0" encoding="UTF-8"?>\n""")
		self.output_file.write("""<sdf3 type="csdf" version="1.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:noNamespaceSchemaLocation="http://www.es.ele.tue.nl/sdf3/xsd/sdf3-csdf.xsd">\n""")
		self.output_file.write("""\t<applicationGraph name=\"""" + name +"""\">\n""")
		self.output_file.write("""\t\t<csdf name=\"""" + name +"""\" type=\"""" + name +"""\">\n""")


	def sdf3_write_actor(self, node):
		"""Writes a normal actor to the XML file"""
		
		self.output_file.write("\n\t\t\t<actor name=\"" + node.name + "\" type=\"Node\">\n")
		for key, val in node.inports.items():
			self.output_file.write("\t\t\t\t<port name=\"p_" + key + "\" type=\"in\" rate=\"")
			for i in range(0,len(val)):
				if i < (len(val) - 1):
					self.output_file.write(val[i] + ",")
				else:
					self.output_file.write(val[i])
			self.output_file.write("\"/>\n")
		for key, val in node.outports.items():
			self.output_file.write("\t\t\t\t<port name=\"p_" + key + "\" type=\"out\" rate=\"")
			for i in range(0,len(val)):
				if i < (len(val) - 1):
					self.output_file.write(val[i] + ",")
				else:
					self.output_file.write(val[i])
			self.output_file.write("\"/>\n")
		self.output_file.write("""\t\t\t\t<port name="_p1" type="out" rate="1"/>\n\t\t\t\t<port name="_p2" type="in" rate="1"/>\n""")
		self.output_file.write("\t\t\t</actor>\n")


	def sdf3_write_actors(self):
		"""Writes all the actors"""
		
		for n in self.nodes_list:
			self.sdf3_write_actor(n)


	def sdf3_write_actors_execution_times(self):
		"""Writes the actors execution time"""
		
		for node in self.nodes_list:
			self.output_file.write("\t\t\t<actorProperties actor=\"" + node.name + "\">\n")
			self.output_file.write("\t\t\t\t<processor type=\"p1\" default=\"true\">\n")
			self.output_file.write("\t\t\t\t\t<executionTime time=\"")
			for i in range(0, len(node.wcet)):
				if i < (len(node.wcet)-1):
					self.output_file.write(node.wcet[i] +",")
				else:
					self.output_file.write(node.wcet[i])
			self.output_file.write("\"/>\n")
			self.output_file.write("\t\t\t\t</processor>\n")
			self.output_file.write("\t\t\t</actorProperties>\n")


	def sdf3_write_channels_sizes(self):
		for e in self.edges_list:
			self.output_file.write("\t\t\t<channelProperties channel=\"" + e.name + "\">\n")
			self.output_file.write("\t\t\t\t<bufferSize sz=\"" + e.size + "\"/>\n")
			self.output_file.write("\t\t\t</channelProperties>\n")


	def sdf3_write_channels(self):
		"""Write all the normal channels"""
		
		i = 1
		for e in self.edges_list:
			self.output_file.write("\t\t\t<channel name=\"" + e.name + "\" srcActor=\"" + self.nodes[e.src_actor].name + "\" srcPort=\"p_" + e.src_port)
			self.output_file.write("\" dstActor=\"" + self.nodes[e.dst_actor].name + "\" dstPort=\"p_" + e.dst_port)
			self.output_file.write("\"/>\n")
			i = i + 1
		

	def sdf3_write_back_edges(self):
		"""
		Writes the back-edges 
		"""
		i = 1
		for n in self.nodes_list:
			self.output_file.write("\t\t\t<channel name=\"_ch" + str(i) + "\" srcActor=\"" + n.name + "\" srcPort=\"_p1\" dstActor=\"" + n.name + "\" dstPort=\"_p2\" initialTokens='1'/>\n")
			i = i + 1


	def sdf3_close_file(self):
		"""
		Closes the file
		"""
		self.output_file.write("\t\t</csdfProperties>\n")
		self.output_file.write("\t</applicationGraph>\n")
		self.output_file.write("</sdf3>\n")


	def read_field(self, field_name):
		line = self.input_file.readline()
		line = line.lstrip('\t')
		s = line.split(":")
		if s[0] != field_name:
			assert False
		return s[1].rstrip('\n')
		
	
	def read_node(self):
		self.read_field("node")

		# read the ID
		node_id = self.read_field("id")

		# read the node name
		node_name = self.read_field("name")

		# read the function name
		function = self.read_field("function")

		# read the sequence length
		seq_length = int(self.read_field("length"))
	   
		# read the WCET
		wcet = self.read_field("wcet").split(" ")

		# read the number of ports
		num_of_ports = int(self.read_field("port_number"))

		in_ports = dict()
		out_ports = dict()
		
		for i in range(0,num_of_ports):
			self.read_field("port")
			port_type = self.read_field("type")
			port_id = self.read_field("id")
			port_rate = self.read_field("rate")
			
			p_rate = port_rate.split(" ")
			assert len(p_rate) == seq_length

			if port_type == "out":
				out_ports[port_id] = p_rate
			elif port_type == "in":
				in_ports[port_id] = p_rate
			else:
				print("Invalid port type!")
				assert False

		node = Node(node_name, function, node_id, seq_length, wcet, in_ports, out_ports)
		self.nodes[node_id] = node
		self.nodes_list.append(node)


	def read_edge(self):
		self.read_field("edge")

		edge_id = self.read_field("id")
		edge_name = self.read_field("name")

		src = self.read_field("src")
		src_s = src.split(" ")
		src_actor = src_s[0]
		src_port = src_s[1]

		dst = self.read_field("dst")
		dst_s = dst.split(" ")
		dst_actor = dst_s[0]
		dst_port = dst_s[1]
		
		ch_name = edge_name
		self.edges_cnt = self.edges_cnt + 1
		edge = Edge(ch_name, edge_id, src_actor, dst_actor, src_port, dst_port)
		self.edges_list.append(edge)


	def extract_repetition_vector(self, child_process):
		# TODO: Use communicate() rather than .stdout
		fp = child_process.stdout
		s = fp.readlines()
		for i in range(1,len(s)-1):
			ss = s[i].decode(ENCODING).split("=")
			n = ss[0].strip()
			nn = n.replace("[","")
			name = nn.replace("]","")
			v = ss[1].strip()
			value = v.replace('\n','')
			self.repetition_vector[name] = value
			

	def sps_write_file(self):
	
		self.output_file.write(str(self.total_nodes) + "\n")
		for n in self.nodes_list:
			self.output_file.write("A " + n.node_id + " " + n.name + " " + n.function + " " + str(n.wcet_val) + " " + self.repetition_vector[n.name] + "\n")
		self.output_file.write(str(self.total_edges) + "\n")
		for e in self.edges_list:
			self.output_file.write("C " + e.edge_id + " " + e.name + " " + e.src_actor + " " + e.dst_actor + " " + str(self.nodes[e.src_actor].seq_length) + " " + str(self.nodes[e.dst_actor].seq_length) + " ")
			for i in self.nodes[e.src_actor].outports[e.src_port]:
				self.output_file.write(i + " ")
			for i in self.nodes[e.dst_actor].inports[e.dst_port]:
				self.output_file.write(i + " ")
			self.output_file.write("\n")


	def generate_files(self, format, name):
		"""Converts .gph to .xml or .sps"""

		self.total_nodes = int(self.read_field("node_number"))
		self.nodes = dict()
		self.nodes_list = []
		for i in range(0,self.total_nodes):
			self.read_node()

		self.total_edges = int(self.read_field("edge_number"))
		self.edges_list = []
		self.edges_cnt = 0
		for i in range(0, self.total_edges):
			self.read_edge()

		sys.stdout.flush()
			
		if format == 0:
			self.sdf3_init_file(name)
			self.sdf3_write_actors()
			self.sdf3_write_channels()
			self.sdf3_write_back_edges()
			self.output_file.write("\t\t</csdf>\n")				
			self.output_file.write("\t\t<csdfProperties>\n")		
			self.sdf3_write_actors_execution_times()
			self.sdf3_close_file()

		elif format == 1:
			self.repetition_vector = dict()
			self.output_file = tempfile.NamedTemporaryFile(mode='w+t', delete=False)
			self.sdf3_init_file(name)
			self.sdf3_write_actors()
			self.sdf3_write_channels()
			self.sdf3_write_back_edges()
			self.output_file.write("\t\t</csdf>\n")				
			self.output_file.write("\t\t<csdfProperties>\n")		
			self.sdf3_write_actors_execution_times()
			self.sdf3_close_file()
			self.output_file.close()

			# Call sdf3analysis-csdf
			child = subprocess.Popen([self.sdf3analysis_path, "--graph", self.output_file.name, "--algo", "repetition_vector"], stdout=subprocess.PIPE)
			self.extract_repetition_vector(child)
			os.unlink(self.output_file.name)
			sys.stdout.flush()
			self.output_file = sys.stdout
			self.sps_write_file()
		else:
			print("Error: Invalid format specified!")
			raise Exception("Invalid format specified")
				
				

def main(name, input_file, output_file, sdf3analysis_path, format):
	x = CSDFConverter(input_file, output_file, sdf3analysis_path)
	x.generate_files(format, name)
	

def print_usage():
	print("""csdfconverter.py [option] < input_file.gph > output_file.(xml,sps)
	Option: 
	-s : specifies to produce the output in SDF^3 XML format. If no option is given, then the output is in SPS format
	""")


if __name__ == '__main__':

	l = len(sys.argv)
	sdf3analysis_tool = "sdf3analysis-csdf"
	input_file = sys.stdin
	output_file = sys.stdout
	name = "myApplication"

	if  l == 2:
		# SDF3 Format
		assert sys.argv[1] == "-s"
		main(name, input_file, output_file, sdf3analysis_tool, 0)
		sys.exit(0)
	elif l == 1:
		# SPS format
		main(name, input_file, output_file, sdf3analysis_tool, 1)		
		sys.exit(0)
	else:
		print_usage()
		sys.exit(-1)

