#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Channel Model for CSDF Graph'

import sys
import math

class ChannelModel:
	"""ChannelModel class
	Represents a communication channel in a CSDF graph
	"""
	
	def __init__(self, channel_id, name, source, destination, production_sequence, consumption_sequence):
		"""Constructor"""
		self.name = name
		self.channel_id = channel_id
		self.source = source
		self.destination= destination
		self.buffer_size = sys.maxsize
		self.production_sequence = production_sequence
		self.consumption_sequence = consumption_sequence
		self.cumulative_prd_rates = []
		self.cumulative_cns_rates = []
		

	def set_source(self, source):
		"""Get the source actor of the channel"""
		self.source = source
		

	def set_destination(self, destination):
		"""Get the destination actor"""
		self.destination = destination
		

	def set_production_sequence(self, production_rate):
		"""Set the production rates sequence"""
		self.production_rate = production_rate
		

	def set_consumption_sequence(self, consumption_sequence):
		"""Set the consumption rates sequence"""
		self.consumption_sequence = consumption_sequence
		

	def set_buffer_size(self, buffer_size):
		"""Set the channel buffer size"""
		assert buffer_size > 0
		self.buffer_size = buffer_size
		

	def set_name(self, name):
		"""Set the name of the channel"""
		self.name = name
		

	def set_cumulative_prd_rates_idp(self, duration, prd_period):
		"""Set the cumulative production rates under the IDP model. Used for computing the earliest start time"""
		firings = int(math.floor(float(duration)/float(prd_period)))
		produced_tokens = 0
		prd_seq = self.get_production_sequence()
		prd_seq_len = len(prd_seq)
		self.cumulative_prd_rates.append(0)
		for i in xrange(0,firings + 1,1):
			produced_tokens = produced_tokens + prd_seq[i % prd_seq_len]
			self.cumulative_prd_rates.append(produced_tokens)
		assert len(self.cumulative_prd_rates) == (firings + 2)
		

	def set_cumulative_prd_rates_cdp(self, duration, prd_period, prd_deadline):
		"""Set the cumulative production rates under the CDP model. Used for computing the earliest start time"""
		firings = int(math.floor(float(duration)/float(prd_period)))			
		if ( float(((duration) % prd_period))/float(prd_deadline)) >= 1.0 :
			firings = firings + 1
		produced_tokens = 0
		prd_seq = self.get_production_sequence()
		prd_seq_len = len(prd_seq)
		self.cumulative_prd_rates.append(0)
		for i in xrange(0,firings + 1,1):
			produced_tokens = produced_tokens + prd_seq[i % prd_seq_len]
			self.cumulative_prd_rates.append(produced_tokens)
		assert len(self.cumulative_prd_rates) == (firings + 2)


	def set_cumulative_cns_rates_cdp(self, duration, cns_period, cns_repetition):
		"""Set the cumulative consumption rates under the CDP model. Used for computing the earliest start time"""
		self.set_cumulative_cns_rates_idp(duration, cns_period, cns_repetition)

	
	def set_cumulative_cns_rates_idp(self, duration, cns_period, cns_repetition):
		"""Set the cumulative consumption rates under the IDP model. Used for computing the earliest start time"""
		firings = int(math.ceil(float(duration)/float(cns_period)))
		if (duration % cns_period == 0 and firings < cns_repetition):
			firings = firings + 1

		consumed_tokens = 0
		cns_seq = self.get_consumption_sequence()
		cns_seq_len = len(cns_seq)	 
		for i in xrange(0,firings + 1,1):   
			consumed_tokens = consumed_tokens + cns_seq[i % cns_seq_len]
			self.cumulative_cns_rates.append(consumed_tokens)


	@staticmethod
	def empty_firings(sequence):
		"""Finds the index of the first zero rate in a prod/cons. sequence"""
		index = -1
		for i in range(0, len(sequence)):
			if sequence[i] > 0:
				index = i
				break
		return index
		
	
	def get_cumulative_prd_rates(self):
		return self.cumulative_prd_rates

	
	def get_cumulative_cns_rates(self):
		return self.cumulative_cns_rates

	
	def clear_cumulative_rates(self):
		self.cumulative_prd_rates = []
		self.cumulative_cns_rates = []
	

	def get_name(self):
		return self.name
		

	def get_channel_id(self):
		return self.channel_id
		

	def get_buffer_size(self):
		return self.buffer_size
		

	def get_source(self):
		return self.source
		

	def get_destination(self):
		return self.destination
		

	def get_production_sequence(self):
		return self.production_sequence
		

	def get_consumption_sequence(self):
		return self.consumption_sequence

	
	def __repr__(self):
		return "Channel ID = " + str(self.channel_id) + ", name = " + self.name + ", source = " + str(self.source) + ", destination = " + str(self.destination) + ", size = " + str(self.buffer_size) + ", production seq. = " + str(self.production_sequence) + ", consumption seq. = " + str(self.consumption_sequence)
		
