#!/usr/bin/python

__author__ =  'Mohamed A. Bamakhrama'
__description__ = 'Platform Generator'

import sys
import os
import math

from ACSDFModel import ACSDFModel
from Mapping import Mapping
from EspamHandler import EspamHandler
from TextHandler import TextHandler
from DotHandler import DotHandler

import Utilities
import PlatformParameters

# Change verbose to True to print detailed information while running
verbose=False

class PlatformGenerator:

	supported_mappings = {	"PEDF-FF" : "Partitioned EDF + first-fit allocation", 
                                "PEDF-FFD" : "Partitioned EDF + first-fit Decreasing allocation", 
				"PEDF-WF" : "Partitioned EDF + worst-fit allocation",
				#"PEDF-WFD" : "Partitioned EDF + worst-fit Decreasing allocation", 
				#"PEDF-BFD" : "Partitioned EDF + best-fit Decreasing allocation", 
				"RM-FF-LB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + first-fit allocation using utilization bound = " + str(Mapping.RM_LOWER_BOUND), 
				"RM-FF-EB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + first-fit allocation using exact utilization bound", 
				"RM-LCE-LB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using exhaustive search and utilization bound = " + str(Mapping.RM_LOWER_BOUND), 
				"RM-LCE-EB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using exhaustive search and exact utilization bound",
				"RM-LCR-LB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using random search and utilization bound = " + str(Mapping.RM_LOWER_BOUND), 
				"RM-LCR-EB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using random search and exact utilization bound",
				"RM-LCG-LB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using genetic algorithm and utilization bound = " + str(Mapping.RM_LOWER_BOUND), 
				"RM-LCG-EB" : "Partitioned Fixed-Priority + Rate-Monotonic priority + low-contention allocation using genetic algorithm and exact utilization bound"
				}


	schedulers_map = {	'PEDF-FF' : Mapping.find_PEDF_FF, 
                                'PEDF-FFD' : Mapping.find_PEDF_FFD, 
				'PEDF-WF' : Mapping.find_PEDF_WF,
				#'PEDF-WFD' : Mapping.find_PEDF_WFD,
				#'PEDF-BFD' : Mapping.find_PEDF_BFD,
  		 		'RM-FF-LB' : Mapping.find_RM_FF,
  		 		'RM-FF-EB' : Mapping.find_RM_FF,
				'RM-LCE-LB' : Mapping.find_RM_LCE,
				'RM-LCE-EB' : Mapping.find_RM_LCE,
				'RM-LCR-LB' : Mapping.find_RM_LCR,
				'RM-LCR-EB' : Mapping.find_RM_LCR,
				'RM-LCG-LB' : Mapping.find_RM_LCG,
				'RM-LCG-EB' : Mapping.find_RM_LCG,
  			}


	def __init__(self):
		self.partitions = []
		self.scheduler_type = "PEDF-FF"
		self.map = None
		self.graphs = []
		self.tasks = []
		self.total_utilization = 0.0
		self.max_utilization = 0.0
		self.total_density = 0.0
		self.max_density = 0.0
		self.num_of_apps = len(self.graphs)
		self.total_utilization = 0.0
		self.beta = 0.0
		self.min_num_of_processors = -1
		
		
	def process(self, scheduler_type, graphs):
			
		self.scheduler_type = scheduler_type
		self.map = self.schedulers_map[scheduler_type]

		assert len(graphs) > 0
		self.graphs = graphs
		self.num_of_apps = len(self.graphs)
		num_of_tasks = 0
		for g in self.graphs:
			for a in g.get_actors():
				a.set_task_id(num_of_tasks)
				self.total_utilization = self.total_utilization + a.get_utilization()
				self.total_density = self.total_density + a.get_density()
				self.tasks.append(a)
				num_of_tasks += 1
				
				if a.get_utilization() > self.max_utilization:
					self.max_utilization = a.get_utilization()
				if a.get_density() > self.max_density:
					self.max_density = a.get_density()
		
		self.beta = int(math.floor(1.0/self.max_utilization))
		
		if self.total_utilization != self.total_density: 
			# CDP
			self.min_num_of_processors = int(math.ceil(self.total_density))
		else: 
			# IDP
			self.min_num_of_processors = int(math.ceil(self.total_utilization))
		

	def get_min_num_of_processors(self):
		return self.min_num_of_processors

	
	def get_num_of_tasks(self):
		return len(self.tasks)
		

	def get_tasks(self):
		return self.tasks
			
			
	def get_scheduler_type(self):
		return self.scheduler_type


	def get_graphs(self):
		return self.graphs
	
	
	def get_num_of_graphs(self):
		return len(self.graphs)
		
	
	def get_task_by_name(self, task_name):
		for t in self.get_tasks():
			if t.get_name() == task_name:
				return t
		return None
	
	def get_partitions(self):
		return self.partitions
	
	def set_partitions(self, partitions):
		self.partitions = partitions
	
	def update_graphs_no_opt(self, mapping):
		amapping = [set() for i in range(len(mapping))]
		j = 0
		for s in mapping:
			for t in s:
				if type(t) == int:
					task = self.get_tasks()[t]
				else:
					task = t
				task.set_affinity(j)
				amapping[j].add(task)
			j += 1
		channels = self.get_mapped_channels(amapping)
		gamma =[[0 for i in range(len(mapping))] for j in range(len(mapping))]
		for c in channels:
			gamma[c[0]][c[1]] = 1

		# Set the contention
		for a in self.get_tasks():
			contention = len(mapping)
			a.set_contention(contention)
			
		for g in self.get_graphs():
			# TODO: This is hard-coded for the time-being ...
			g.update_parameters(PlatformParameters.AXI_CB, PlatformParameters.T_RD_AXI_CB, PlatformParameters.T_WR_AXI_CB)
			
	
	def update_graphs(self, mapping):
		amapping = [set() for i in range(len(mapping))]
		j = 0
		for s in mapping:
			for t in s:
				if type(t) == int:
					task = self.get_tasks()[t]
				else:
					task = t
				task.set_affinity(j)
				amapping[j].add(task)
			j += 1
		channels = self.get_mapped_channels(amapping)
		gamma =[[0 for i in range(len(mapping))] for j in range(len(mapping))]
		for c in channels:
			gamma[c[0]][c[1]] = 1

		# Set the contention
		for a in self.get_tasks():
			contention = sum(gamma[a.get_affinity()])
			a.set_contention(contention)
			
		for g in self.get_graphs():
			g.update_parameters(PlatformParameters.AXI_CB, PlatformParameters.T_RD_AXI_CB, PlatformParameters.T_WR_AXI_CB)


	def get_mapped_channels(self,partitions):

		channels = []
		chan_cnt = 0
		for a in self.get_tasks():
			actor = a
			for outport in actor.get_outports():
				ch = actor.get_graph().get_channel(outport)
				successor = actor.get_graph().get_actor(ch.get_destination())				
				channels.append((actor.get_affinity(), successor.get_affinity()))
				chan_cnt += 1
		return channels


def print_usage():
	sys.stderr.write("""
Usage:
\t> PlatformGenerator.py out_dir time_in_cycles map_type g_1_f g_1_ts g_1.csf g_2_f g_2_ts g_2.csf ...
where:
\t- out_dir: is the directiry where the platform and mapping specifications will be written
\t- time_in_cycles: is a binary flag (1 or 0) which specifies whether the time should be shown in cycles or OS clock ticks
\t- map_type: is the type of mapping. Currently, the supported types are:
""")
	for e in PlatformGenerator.supported_mappings:
		sys.stderr.write("\t\t* " + e + " : " + PlatformGenerator.supported_mappings[e] + "\n")
	sys.stderr.write(
"""\t- g_1_f: graph g_1 period scaling factor
\t- g_1_ts: graph g_1 maximum token size
\t- g_1.csf: path to graph g_1 CSDF graph file in the Compact StreamIt format\n\n""")   


def make_platform(output_dir, scheduler_type, factors, token_sizes, apps, printinfo=False):
	
	assert len(factors) == len(token_sizes) == len(apps)

	graphs = list()
	
	for i in range(0, len(apps)):

		g = ACSDFModel.load_graph_from_file(apps[i], 1)

		if printinfo:
			sys.stdout.write("\n")
			sys.stdout.write(" - Processing application " + os.path.basename(apps[i]) + "...\n")		
			sys.stdout.write(" -- Number of actors = " + str(g.get_num_of_actors()) + "\n")
			sys.stdout.write(" -- Repetition vector = " + g.get_repetition_vector_str() + "\n")
			sys.stdout.write(" -- Worst-case execution time vector = " + g.get_execution_vector_str() + "\n")
			sys.stdout.write(" -- Minimum period vector = " + g.get_period_vector_str() + "\n")
			sys.stdout.write(" -- Deadline vector = " + g.get_deadline_vector_str() + "\n")
            
		g.set_factor(factors[i])
		g.set_max_token_size(token_sizes[i])
		
		g.update_parameters(PlatformParameters.AXI_CB, PlatformParameters.T_RD_AXI_CB, PlatformParameters.T_WR_AXI_CB)
		g.calc_start_time_buffer_vector(ACSDFModel.IDP_MIN, 0)

		if verbose:
			sys.stdout.write("\n")
			sys.stdout.write(" -- Least common multiple of the repetitions (Q) = " + str(g.lcm) + "\n") 
			sys.stdout.write(" -- max(WCET_i * q_i) (eta) = " + str(g.max_q_C) + "\n") 
			sys.stdout.write(" -- Iteration period = " + str(g.alpha) + "\n")

			if g.get_matched_io_type():
				sys.stdout.write(" -- " + g.get_graph_name() + " is a matched I/O rates graph\n")
			else:
				sys.stdout.write(" -- " + g.get_graph_name() + " is a mis-matched I/O rates graph\n")

			if g.get_minor_cycle() != 0:
				sys.stdout.write(" -- Cyclic executive major cycle = " + str(g.get_major_cycle()) + ", Cyclic Executive Minor Cycle = " + str(g.get_minor_cycle()) + "\n")
			else:
				sys.stdout.write(" -- Unable to find a minor cycle with no preemptions!\n")

			sys.stdout.write(" -- Hyperperiod = " + str(Utilities.lcmv(g.get_period_vector())) + "\n")

		graphs.append(g)
		
		#ACSDFModel.print_graph_dot(g, output_dir)
	if printinfo:	
		sys.stdout.write("\n")
	
	pg = PlatformGenerator()
	pg.process(scheduler_type, graphs)

	# Mapping phase
	
	if pg.get_scheduler_type().startswith("RM"):
		if pg.get_scheduler_type().endswith("LB"):
			partitions = pg.map(pg, Mapping.LOWER_BOUND)
		elif pg.get_scheduler_type().endswith("EB"):
			partitions = pg.map(pg, Mapping.EXACT_BOUND)
		else:
			assert False
	else:	
		partitions = pg.map(pg)

	# In case of exhaustive search, the results is actually a list of the best mappings
	if pg.get_scheduler_type().startswith("RM-LCE"):
		partitions = partitions[-1] 
		
	pg.update_graphs(partitions)
	
	pg.set_partitions(partitions)
	
	if verbose:
		sys.stdout.write(" - Contention of the mapping = " + str(Mapping.contention(partitions,pg)) + "\n")
		sys.stdout.write(" - Mapping is " + str(partitions))
		sys.stdout.write("\n\n")
			
	# Update start times and buffer sizes
	for g in pg.get_graphs():
		g.calc_start_time_buffer_vector(ACSDFModel.IDP_MIN, 0)
		if printinfo:
			sys.stdout.write(" - Dumping " + g.get_graph_name() + " information...\n")
			sys.stdout.write(" -- Period vector = " + g.get_period_vector_str() + "\n")
			sys.stdout.write(" -- Start time vector = " + g.get_start_time_vector_str() + "\n")
			sys.stdout.write(" -- Buffer size vector = " + g.get_buffer_size_vector_str() + "\n")
		
			sys.stdout.write(" -- Sink actors periods:")
			g_levels = g.find_levels()
			for aa in g_levels[-1]:
				ac = g.get_actor(aa.get_actor_id())
				sys.stdout.write(" " + ac.get_full_name() + " (" + ac.get_function() + ") : " + str(ac.get_period()) + "\n")
			
			sys.stdout.write(" -- Total Utilization  = " + str(g.get_utilization()) + "\n")
			sys.stdout.write(" -- Graph maximum latency under the implicit-deadline model = " + str(g.get_idp_latency()) + "\n")
			sys.stdout.write("\n")

	if printinfo:
		sys.stdout.write(" - Minimum number of processors needed to schedule the applications by:\n") 
		sys.stdout.write(" -- Optimal scheduling algorithms (M_OPT) = " + str(pg.get_min_num_of_processors()) + "\n")
		sys.stdout.write(" -- " + PlatformGenerator.supported_mappings[pg.get_scheduler_type()] + " (M_" + pg.get_scheduler_type() + ") = " + str(len(partitions)) + "\n")
		sys.stdout.write("\n")
		
	# Call ESPAM handler
	eh = EspamHandler(pg, 'ML605', output_dir, partitions)
	eh.generate_platform_file()
	eh.generate_mapping_file()

	# Call text handler
	th = TextHandler(pg, output_dir, partitions)
	th.generate_mapping_file()

	dh = DotHandler(pg, output_dir, partitions)
	dh.generate_mapping_file()
	return pg

def main():
	APP_START_INDEX = 4
	if len(sys.argv) < (APP_START_INDEX + 3) or (len(sys.argv) - APP_START_INDEX) % 3 != 0:
		print_usage()
		sys.exit(-1)
		
	output_dir = sys.argv[1]
	time_in_cycles = int(sys.argv[2])
	scheduler_type = sys.argv[3]
	
	if not os.path.isdir(output_dir):
		sys.stderr.write("ERROR! " + output_dir + " is not a valid directory \n")
		print_usage()
		sys.exit(-1)
	
	if time_in_cycles != 1 and time_in_cycles != 0:
		sys.stderr.write("ERROR! Invalid value of time units \n")
		print_usage()
		sys.exit(-1)

	if scheduler_type not in PlatformGenerator.supported_mappings:
		sys.stderr.write("ERROR! Unsupported mapping \n")
		print_usage()
		sys.exit(-1)

	sys.stdout.write("***************************************************************************\n")
	sys.stdout.write("*   csdf-rtschedtools: A toolbox for hard-real-time multiprocessor        *\n")
	sys.stdout.write("*   schedulability analysis of acyclic Cyclo-Static Dataflow graphs       *\n")
	sys.stdout.write("*   Copyright (c) 2012 by Leiden University. All rights reserved.         *\n")
	sys.stdout.write("*   See LICENSE.md for license information                                *\n")
	sys.stdout.write("***************************************************************************\n")

	# Check that we have proper number of application parameters
	assert (len(sys.argv) - APP_START_INDEX) % 3 == 0

	factors = []
	token_sizes = []
	apps = []

	for i in range(APP_START_INDEX, len(sys.argv), 3):
		factors.append(int(sys.argv[i]))
		token_sizes.append(int(sys.argv[i + 1]))
		apps.append(sys.argv[i + 2])

	if time_in_cycles == 0:
		PlatformParameters.TIME_IN_CYCLES = False

	make_platform(output_dir, scheduler_type, factors, token_sizes, apps, printinfo=True)
        
        
if __name__ == "__main__":
	main()
		
